import torch
from torch.nn.functional import cross_entropy, one_hot, softmax


def seesaw_ce_loss(pred,
                   true,
                   cum_samples,
                   num_classes,
                   p, q, eps):
    assert pred.size(-1) == num_classes
    assert len(cum_samples) == num_classes

    onehot_labels = one_hot(true, num_classes)
    seesaw_weights = pred.new_ones(onehot_labels.size())

    # mitigation factor
    if p > 0:
        sample_ratio_matrix = cum_samples[None, :].clamp(min=1) / cum_samples[:, None].clamp(min=1)
        index = (sample_ratio_matrix < 1.0).float()
        sample_weights = sample_ratio_matrix.pow(p) * index + (1 - index)  # M_{ij}
        mitigation_factor = sample_weights[true.long(), :]
        seesaw_weights = seesaw_weights * mitigation_factor

    # compensation factor
    if q > 0:
        scores = softmax(pred.detach(), dim=1)
        self_scores = scores[torch.arange(0, len(scores)).to(scores.device).long(), true.long()]
        score_matrix = scores / self_scores[:, None].clamp(min=eps)
        index = (score_matrix > 1.0).float()
        compensation_factor = score_matrix.pow(q) * index + (1 - index)
        seesaw_weights = seesaw_weights * compensation_factor

    pred = pred + (seesaw_weights.log() * (1 - onehot_labels))

    return cross_entropy(pred, true, weight=None, reduction='none')


class SeesawLoss(torch.nn.Module):
    """
    Implementation of seesaw loss.
    Refers to `Seesaw Loss for Long-Tailed Instance Segmentation (CVPR 2021)
    <https://arxiv.org/abs/2008.10032>`_
    Args:
        num_classes (int):
         The number of classes.
         Default to 1000 for the ImageNet dataset.
        p (float):
         The ``p`` in the mitigation factor.
         Defaults to 0.8.
        q (float):
         The ``q`` in the compensation factor.
         Defaults to 2.0.
        eps (float):
         The min divisor to smooth the computation of compensation factor.
         Default to 1e-2.
    """

    def __init__(self, num_classes=1000,
                 p=0.8, q=2.0, eps=1e-2):
        super(SeesawLoss, self).__init__()
        self.p = p
        self.q = q
        self.eps = eps
        self.num_classes = num_classes

        # cumulative samples for each category
        self.register_buffer('cum_samples',
                             torch.zeros(self.num_classes, dtype=torch.float))

    def forward(self, pred, true):
        # accumulate the samples for each category
        unique_labels = true.unique()
        for unique in unique_labels:
            self.cum_samples[unique] += (true == unique.item()).sum()

        return seesaw_ce_loss(pred, true,
                              self.cum_samples,
                              self.num_classes,
                              self.p, self.q, self.eps)
